/* -*- Mode: C++; tab-width: 20; indent-tabs-mode: nil; c-basic-offset: 4 -*-
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is Netscape Communications
 * Corporation. Portions created by the Initial Developer are
 * Copyright (C) 2003 the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *    Ralph Thomas <rthomas@yahoo-inc.com>
 *    Peter Amstutz <tetron@interreality.org>
 *    Stuart Parmenter <pavlov@netscape.com>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */

#include "membufWindow.h"
#include "nsGUIEvent.h"
#include "nsIRenderingContext.h"
#include "nsIThebesRenderingContext.h"
#include "nsIDeviceContext.h"
#include "nsIComponentManager.h"
#include "nsGfxCIID.h"
#include "nsIRegion.h"
#include "plevent.h"
#include "prlog.h"
#include "nsIServiceManager.h"
#include "nsIEventQueueService.h"
#include "gfxPlatformMembuf.h"
#include "gfxImageSurface.h"

#include "prlog.h"
#include "prmem.h"
#include "prenv.h"

#ifdef WIN32
// Required for GetMessageTime.
#include <windows.h>
#include <stdio.h>
#endif


#define MEMBUF_NATIVE_SURFACE		7703
#define MEMBUF_NATIVE_REGION		7704
#define MEMBUF_NATIVE_DATA		7705
#define MEMBUF_NATIVE_WIDTH		7706
#define MEMBUF_NATIVE_HEIGHT		7707

static NS_DEFINE_IID(kRenderingContextCID, NS_RENDERING_CONTEXT_CID);
static NS_DEFINE_CID(kEventQueueServiceCID, NS_EVENTQUEUESERVICE_CID);
static NS_DEFINE_IID(kDeviceContextCID, NS_DEVICE_CONTEXT_CID);
static NS_DEFINE_CID(kRegionCID, NS_REGION_CID);

NS_IMPL_ISUPPORTS_INHERITED0(membufWindow, nsBaseWidget)

#ifdef PR_LOGGING
PRLogModuleInfo* nsMembufWindowLog;
#endif

//
// This membuf implementation uses Cairo/Thebes for the graphics
// layer.
//

membufWindow::membufWindow() :
  mRootSurface(0), mIsVisible(PR_FALSE), mCreated(PR_FALSE)
{
#ifdef PR_LOGGING
    if (!nsMembufWindowLog)
        nsMembufWindowLog = PR_NewLogModule ("nsMembufWindowLog");
    PR_LOG(nsMembufWindowLog, PR_LOG_DEBUG, ("(0x%x) new membufWindow", this));
#endif
    mBounds = nsRect(0, 0, 800, 600);
    mDirtyRegion = do_CreateInstance(kRegionCID);
}

membufWindow::~membufWindow()
{
    if (mRootSurface) mRootSurface = 0;
#ifdef PR_LOGGING
    PR_LOG(nsMembufWindowLog, PR_LOG_DEBUG, ("(0x%x) delete membufWindow", this));
#endif
}

//////////////////////////////////////////////
//// nsIWidget

//
// The root membufWindow allocates a surface in GetThebesSurface.
// All of it's child windows draw into this surface, and use the
// gfxASurface::SetDeviceOffset mechanism.
//
// In this way we only ever allocate one surface. The surface is
// reallocated every time the widget is resized, and as
// GetThebesSurface knows to ask the parent for it's surface each
// time (if the window is not a root) we don't have to keep
// any refcounts around, as the original membuf code did.
//

NS_IMETHODIMP
membufWindow::Create(nsIWidget *aParent,
                   const nsRect     &aRect,
                   EVENT_CALLBACK   aHandleEventFunction,
                   nsIDeviceContext *aContext,
                   nsIAppShell      *aAppShell,
                   nsIToolkit       *aToolkit,
                   nsWidgetInitData *aInitData)
{
    mCreated = PR_TRUE;
    nsIWidget *baseParent = aInitData &&
                 (aInitData->mWindowType == eWindowType_dialog ||
                  aInitData->mWindowType == eWindowType_toplevel) ?
                  nsnull : aParent;

    nsCOMPtr<nsIDeviceContext> dc( aContext );
    if (dc) dc->Init(nsnull);
    else {
	nsCOMPtr<nsIDeviceContext> dc = do_CreateInstance(kDeviceContextCID);
	dc->Init(nsnull);
    }

    BaseCreate(baseParent, aRect, aHandleEventFunction, aContext,
               aAppShell, aToolkit, aInitData);
    mBounds = aRect;
    mParent = baseParent;
#ifdef PR_LOGGING
    PR_LOG(nsMembufWindowLog, PR_LOG_DEBUG, ("(0x%x) membufWindow::Create[ (nsIWidget) baseParent = 0x%x]", this, baseParent));
#endif
    return NS_OK;
}

NS_IMETHODIMP
membufWindow::Create(nsNativeWidget aParent,
                   const nsRect     &aRect,
                   EVENT_CALLBACK   aHandleEventFunction,
                   nsIDeviceContext *aContext,
                   nsIAppShell      *aAppShell,
                   nsIToolkit       *aToolkit,
                   nsWidgetInitData *aInitData)
{
    mCreated = PR_TRUE;

    nsCOMPtr<nsIDeviceContext> dc( aContext );

    if (dc) dc->Init(0);
    else {
	dc = do_CreateInstance(kDeviceContextCID);
	dc->Init(nsnull);
    }

    BaseCreate(nsnull, aRect, aHandleEventFunction, aContext,
               aAppShell, aToolkit, aInitData);

    mBounds = aRect;

    //
    // This constructor shouldn't really be called, unless it is
    // intended that this widget be a root. This gets called when
    // embedding...
    //
#ifdef PR_LOGGING
    PR_LOG(nsMembufWindowLog, PR_LOG_DEBUG, ("(0x%x) membufWindow::Create[ (nsNativeWidget) baseParent = 0x0]", this));
#endif

    return NS_OK;
}

void membufWindow::calculateAbsolutePosition(int& total_offset_x, int& total_offset_y)
{
    total_offset_x = mBounds.x;
    total_offset_y = mBounds.y;

    for(nsIWidget* par = mParent; par->GetParent(); par = par->GetParent()) {
        nsRect r;
        par->GetBounds(r);
        total_offset_x += r.x;
        total_offset_y += r.y;
    }
#ifdef PR_LOGGING
    PR_LOG(nsMembufWindowLog, PR_LOG_DEBUG, ("(0x%x) membufWindow::calculateAbsolutePosition[ %d+%d ]", this, total_offset_x, total_offset_y));
#endif

}

NS_IMETHODIMP
membufWindow::Destroy(void)
{
    nsBaseWidget::Destroy();
    mParent = nsnull;
    mEventCallback = nsnull;
    mCreated = PR_FALSE;

    if( mRootSurface ) mRootSurface = 0;

#ifdef PR_LOGGING
    PR_LOG(nsMembufWindowLog, PR_LOG_DEBUG, ("(0x%x) membufWindow::Destroy[]", this));
#endif

    return NS_OK;
}

NS_IMETHODIMP
membufWindow::SetParent(nsIWidget* aNewParent)
{
    mParent = aNewParent;
#ifdef PR_LOGGING
    PR_LOG(nsMembufWindowLog, PR_LOG_DEBUG, ("(0x%x) membufWindow::SetParent[ aNewParent = 0x%x]", this, aNewParent));
#endif

    return NS_OK;
}

nsIWidget*
membufWindow::GetParent(void)
{
    nsIWidget *ret;
    ret = mParent;
    NS_IF_ADDREF(ret);

#ifdef PR_LOGGING
    PR_LOG(nsMembufWindowLog, PR_LOG_DEBUG, ("(0x%x) membufWindow::GetParent[ 0x%x ]", this, ret));
#endif

    return ret;
}

NS_IMETHODIMP
membufWindow::Show(PRBool aState)
{
    mIsVisible = aState;
    if(aState) Invalidate(PR_FALSE);
    else {
        nsIWidget* par = this;
        while(par->GetParent()) par = par->GetParent();
        if(par == this) return NS_OK;
        else {
            par->Invalidate(PR_FALSE);
        }
    }

#ifdef PR_LOGGING
    PR_LOG(nsMembufWindowLog, PR_LOG_DEBUG, ("(0x%x) membufWindow::Show[ %d ]", this, aState));
#endif

    return NS_OK;
}

NS_IMETHODIMP
membufWindow::SetModal(PRBool aModal)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
membufWindow::IsVisible(PRBool & aState)
{
    aState = mIsVisible;
    return NS_OK;
}

NS_IMETHODIMP
membufWindow::ConstrainPosition(PRBool aAllowSlop,
                              PRInt32 *aX,
                              PRInt32 *aY)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
membufWindow::Move(PRInt32 aX, PRInt32 aY)
{
    return Resize(aX, aY, mBounds.width, mBounds.height, 0);
}

NS_IMETHODIMP
membufWindow::Resize(PRInt32 aWidth,
                   PRInt32 aHeight,
                   PRBool  aRepaint)
{
    return Resize(mBounds.x, mBounds.y, aWidth, aHeight, aRepaint);
}

NS_IMETHODIMP
membufWindow::Resize(PRInt32 aX,
                   PRInt32 aY,
                   PRInt32 aWidth,
                   PRInt32 aHeight,
                   PRBool aRepaint)
{
    mBounds = nsRect(aX, aY, aWidth, aHeight);
    if(aRepaint) {
        mDirtyRegion->Union(aX, aY, aWidth, aHeight);
        InvalidateHelper(false);
    }
    

#ifdef PR_LOGGING
    PR_LOG(nsMembufWindowLog, PR_LOG_DEBUG, ("(0x%x) membufWindow::Resize[ aX = %d, aY = %d, aWidth = %d, aHeight = %d, aRepaint = %d ]", this, aX, aY, aWidth, aHeight, aRepaint));
#endif


    return NS_OK;
}

NS_IMETHODIMP
membufWindow::PlaceBehind(nsTopLevelWidgetZPlacement aPlacement,
                          nsIWidget *aWidget, PRBool aActivate)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
membufWindow::Enable(PRBool aState)
{
//    mEnabled = aState;
    return NS_OK;
}

NS_IMETHODIMP
membufWindow::IsEnabled(PRBool *aState)
{
    *aState = PR_TRUE;//mEnabled;
    return NS_OK;
}

// Focus (mostly) works as well.

NS_IMETHODIMP
membufWindow::SetFocus(PRBool aRaise)
{
    /*if(!hasFocus)*/ DispatchGotFocusEvent();

    return NS_OK;
}

NS_IMETHODIMP
membufWindow::LoseFocus()
{
    /*if(hasFocus)*/ DispatchLostFocusEvent();

    return NS_OK;
}

NS_IMETHODIMP
membufWindow::GetBounds(nsRect &aRect)
{
    aRect = mBounds;
    return NS_OK;
}

NS_IMETHODIMP
membufWindow::GetScreenBounds(nsRect &aRect)
{
    return GetBounds(aRect);
}

NS_IMETHODIMP
membufWindow::GetClientBounds(nsRect &aRect)
{
    return GetBounds(aRect);
}

NS_IMETHODIMP
membufWindow::GetBorderSize(PRInt32 &aWidth, PRInt32 &aHeight)
{
    aWidth = aHeight = 0;
    return NS_OK;
}

nsIFontMetrics* membufWindow::GetFont(void)
{
    return nsnull;
}

NS_IMETHODIMP
membufWindow::SetFont(const nsFont &aFont)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

nsCursor membufWindow::GetCursor(void)
{
    return mCursor;
}

NS_IMETHODIMP
membufWindow::SetCursor(nsCursor aCursor)
{
    mCursor = aCursor;
    return NS_OK;
}

NS_IMETHODIMP
membufWindow::Validate()
{
    mDirtyRegion->SetTo(0, 0, 0, 0);
    return NS_OK;
}

NS_IMETHODIMP
membufWindow::Invalidate(PRBool aIsSynchronous)
{
    Invalidate(mBounds, aIsSynchronous);

    return NS_OK;
}

NS_IMETHODIMP
membufWindow::Invalidate(const nsRect & aRect, PRBool aIsSynchronous)
{
    mDirtyRegion->Union(aRect.x, aRect.y, aRect.width, aRect.height);
    InvalidateHelper(aIsSynchronous);

    return NS_OK;
}

NS_IMETHODIMP
membufWindow::InvalidateRegion(const nsIRegion* aRegion, PRBool aIsSynchronous)
{
    mDirtyRegion->Union(*aRegion);
    InvalidateHelper(aIsSynchronous);

    return NS_OK;
}

NS_IMETHODIMP
membufWindow::Update()
{
    //
    // For testing purposes, if the MOZ_MEMBUF_DUMP_PPM environment variable
    // is set, we will dump out a PPM of the framebuffer every time we do
    // some painting in this function.
    //
    static PRBool dumpGraphics = PR_GetEnv("MOZ_MEMBUF_DUMP_PPM") != nsnull;
    PRBool didPaint = false;
#ifdef PR_LOGGING
    PR_LOG(nsMembufWindowLog, PR_LOG_DEBUG, ("(0x%x) membufWindow::Update[]", this));
#endif
    if (!mCreated) {
#ifdef PR_LOGGING
        PR_LOG(nsMembufWindowLog, PR_LOG_DEBUG, ("(0x%x) membufWindow::Update -- FAILED: window not created yet, or has been destroyed.", this));
#endif
        return NS_OK;
    }
    //
    // Redraw all dirty rectangles. I don't know if GetRects is returning
    // an optimized list of rectangles or not...
    //
    if(! mDirtyRegion->IsEmpty()) {
//        nsRegionRectSet* _rrs = 0;
//        mDirtyRegion->GetRects(&_rrs);
        nsIRenderingContext* rc = GetRenderingContext();
	//
	// We set the clipping so that a child window can't destroy
        // the contents of the framebuffer outside of it's bounds.
        //
        // XXX: I'm still having trouble with iframes, especially
        // XXX: those for Google Ads. I don't think that this clipping
        // XXX: fully works -- or there's another issue that I'm not
        // XXX: aware of. <URL:http://www.digg.com/> showcases this
        // XXX: issue well.
        //
        rc->SetClipRect( mBounds, nsClipCombine_kReplace );
	rc->PushState();
//        for(PRUint32 _i = 0; _i < _rrs->mNumRects; _i++) {
            Paint(*rc, mDirtyRegion );
//        }
//        mDirtyRegion->FreeRects(_rrs);
        rc->PopState();
        NS_RELEASE(rc);
        didPaint = true;
        //
        // Update the cumulative region. If we don't have a cumulative
        // region then create one.
        //
        if (!mCumulativeRegion)
        {
            mCumulativeRegion = do_CreateInstance(kRegionCID);
            mCumulativeRegion->Init();
        }
        mCumulativeRegion->Union(*mDirtyRegion);
    }
    Validate();
    //
    // Now that painting has apparently taken place, try dumping the result
    // out to a PPM.
    //
    if( dumpGraphics && didPaint )
    {
        static unsigned int count = 0;
#ifdef PR_LOGGING
        PR_LOG(nsMembufWindowLog, PR_LOG_DEBUG, ("(0x%x) membufWindow::Update -- dumping PPM", this));
#endif
        count++;
        nsCAutoString name(PR_GetEnv("MOZ_MEMBUF_DUMP_PPM"));
        name.AppendLiteral( "-" );
        name.AppendInt( count );
        name.AppendLiteral( ".ppm" );

        gfxImageSurface* imgSurf = (gfxImageSurface*)GetThebesSurface();
        if (!imgSurf) return NS_OK;
        imgSurf->Flush();
        FILE* f = fopen(name.get(), "wb");
        if (!f) return NS_OK;
        fprintf(f, "P6\n%d\n%d\n255\n", imgSurf->Width(), imgSurf->Height());
        unsigned char* row = new unsigned char[3 * imgSurf->Width()];
        for(PRInt32 y = 0; y < imgSurf->Height(); y++)
        {
            unsigned char* src = imgSurf->Data() + (y*imgSurf->Width()*4);
            unsigned char* dst = row;
            for(PRInt32 x = 0; x < imgSurf->Width(); x++)
            {
                dst[0] = src[2];
                dst[1] = src[1];
                dst[2] = src[0];
                dst += 3; src += 4;
            }
            fwrite(row, 3, imgSurf->Width(), f);
        }
        fclose(f);
        delete[] row;
    }
    //
    // Notify any observers if we painted.
    //
    if (didPaint)
    {
        nsresult rv = 0;
        nsCOMPtr<nsIObserverService> observerService = do_GetService("@mozilla.org/observer-service;1", &rv);
        nsISupports *subject = (nsISupports*)((nsIWidget*)this);
        nsAFlatString context(NS_LITERAL_STRING("update"));
        observerService->NotifyObservers(subject, "window-update",
            context.get());
    }

    return NS_OK;
}

NS_IMETHODIMP
membufWindow::SetColorMap(nsColorMap *aColorMap)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}


NS_IMETHODIMP
membufWindow::ScrollWidgets(PRInt32 aDx, PRInt32 aDy)
{
    if(aDx == 0 && aDy == 0) return NS_OK;

    //
    // XXXralpht: For now we just redraw the entire viewport. Eventually
    // XXXralpht: I will make it perform copying instead of redrawing.
    //
    // Depending on if the window that recieves this call is a root or
    // not (almost certainly not) this can be a pretty simple memcpy, or
    // a slightly more complicated row copy.
    //

    Invalidate(PR_TRUE);
    return NS_OK;
}

NS_IMETHODIMP
membufWindow::Scroll(PRInt32 aDx, PRInt32 aDy, nsRect* aClipRect)
{
    Invalidate( PR_FALSE );
    if (!mRootSurface) return NS_OK;
    //
    // We need to scroll any child windows.
    //
    for(nsIWidget* wid = GetFirstChild(); wid; wid = wid->GetNextSibling())
    {
        nsRect r;
        wid->GetBounds(r);
	r.x += aDx; r.y += aDy;
        NS_STATIC_CAST(nsBaseWidget*, wid)->SetBounds(r);
    }
    return ScrollWidgets(aDx, aDy);
}

NS_IMETHODIMP
membufWindow::ScrollRect(nsRect &aSrcRect, PRInt32 aDx, PRInt32 aDy)
{
    return Scroll(aDx, aDy, 0);
}

// This is the hook by which the application can access the bitmap for
// this window, the topmost bitmap (the parent window if this is a
// sub-bitmap), the reference count (so that if a Resize() occurs the
// application doesn't find the bitmap it is reading from suddenly
// deleted) and a dirty flag which can be used by the application to
// efficiently check whether the underlying bitmap has changed and
// needs to be refreshed (if for example the application copies the
// bitmap buffer itself).

void* membufWindow::GetNativeData(PRUint32 aDataType)
{
    switch(aDataType) {
    case MEMBUF_NATIVE_SURFACE:
    case NS_NATIVE_WINDOW: {
        gfxImageSurface* surf = (gfxImageSurface*)mRootSurface.get();
        if (surf) printf("membuf: %dx%d\n", surf->Width(), surf->Height());
        return (void*)surf;
    }

    case MEMBUF_NATIVE_REGION: {
        //
        // Wipe out any dirty cumulative region that we have and return
        // it. This assumes that whoever gets this cumulative region is
        // going to do something useful with it.
        //
        if (!mCumulativeRegion) return 0;
        nsIRegion* dirty = mCumulativeRegion.get();
        NS_ADDREF(dirty);
        mCumulativeRegion = 0;
        return dirty;
    }

    case MEMBUF_NATIVE_DATA: {
        gfxImageSurface* surf = (gfxImageSurface*)mRootSurface.get();
        if (surf) return surf->Data();
        return 0;
    }

    case MEMBUF_NATIVE_WIDTH: {
        gfxImageSurface* surf = (gfxImageSurface*)mRootSurface.get();
        if (surf) return (void*)surf->Width();
        return 0;
    }

    case MEMBUF_NATIVE_HEIGHT: {
        gfxImageSurface* surf = (gfxImageSurface*)mRootSurface.get();
        if (surf) return (void*)surf->Height();
        return 0;
    }

    case NS_NATIVE_GRAPHIC:
    case NS_NATIVE_COLORMAP:
    case NS_NATIVE_WIDGET:
    case NS_NATIVE_DISPLAY:
    case NS_NATIVE_REGION:
    case NS_NATIVE_OFFSETX:
    case NS_NATIVE_OFFSETY:
    case NS_NATIVE_PLUGIN_PORT:
    case NS_NATIVE_SCREEN:
    default:
        return 0;
    }
}
void membufWindow::FreeNativeData(void * data, PRUint32 aDataType)
{
}

NS_IMETHODIMP
membufWindow::SetTitle(const nsAString& aTitle)
{
    return NS_OK;
}

NS_IMETHODIMP
membufWindow::SetIcon(const nsAString& anIconSpec)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
membufWindow::SetMenuBar(nsIMenuBar * aMenuBar)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}
NS_IMETHODIMP
membufWindow::ShowMenuBar(PRBool aShow)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
membufWindow::WidgetToScreen(const nsRect& aOldRect, nsRect& aNewRect)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
membufWindow::ScreenToWidget(const nsRect& aOldRect, nsRect& aNewRect)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
membufWindow::BeginResizingChildren(void)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
membufWindow::EndResizingChildren(void)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
membufWindow::GetPreferredSize(PRInt32& aWidth, PRInt32& aHeight)
{
    aWidth = 800;
    aHeight = 600;
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
membufWindow::SetPreferredSize(PRInt32 aWidth, PRInt32 aHeight)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

nsIWidget* findWidgetAt(nsIWidget* start, int x, int y)
{
    for(nsIWidget* wid = start->GetFirstChild();
        wid != 0;
        wid = wid->GetNextSibling())
    {
        nsRect r;
        wid->GetBounds(r);

        if(x >= r.x && y >= r.y && x <= (r.x + r.width) && y <= (r.y + r.height))
            {
            return findWidgetAt(wid, x - r.x, y - r.y);
        }
    }

    return start;
}

NS_IMETHODIMP
membufWindow::DispatchEvent(nsGUIEvent* aEvent, nsEventStatus & aStatus)
{
    aStatus = nsEventStatus_eIgnore;

    if(aEvent->eventStructType == NS_MOUSE_EVENT) {
        nsMouseEvent* me = (nsMouseEvent*)aEvent;

        if(aEvent->message == NS_MOUSE_LEFT_CLICK) {
//            if(!hasFocus) {
                DispatchActivateEvent();
                SetFocus(PR_TRUE);
//            }
        }

        /*
        nsIWidget* wid = findWidgetAt(this, me->point.x, me->point.y);
        nsRect r;
        wid->GetBounds(r);
        //printf("inside widget %x with bounds %i %i %i %i\n", wid, r.x, r.y, r.width, r.height);
        */
    }

    // hold a widget reference while we dispatch this event
    NS_ADDREF(aEvent->widget);

    // send it to the standard callback
    if (mEventCallback)
        aStatus = (* mEventCallback)(aEvent);

    // dispatch to event listener if event was not consumed
    if ((aStatus != nsEventStatus_eIgnore) && mEventListener)
        aStatus = mEventListener->ProcessEvent(*aEvent);

    NS_IF_RELEASE(aEvent->widget);

    return NS_OK;
}

NS_IMETHODIMP
membufWindow::EnableDragDrop(PRBool aEnable)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

void
membufWindow::ConvertToDeviceCoordinates(nscoord &aX,nscoord &aY)
{
}

NS_IMETHODIMP
membufWindow::CaptureMouse(PRBool aCapture)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
membufWindow::GetWindowClass(char *aClass)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
membufWindow::SetWindowClass(char *aClass)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
membufWindow::CaptureRollupEvents(nsIRollupListener * aListener, PRBool aDoCapture, PRBool aConsumeRollupEvent)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
membufWindow::ModalEventFilter(PRBool aRealEvent, void *aEvent, PRBool *aForWindow)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

NS_IMETHODIMP
membufWindow::GetAttention(PRInt32 aCycleCount)
{
    return NS_ERROR_NOT_IMPLEMENTED;
}

//////////////////////////////////////////////
//// membufWindow protected

struct AsyncUpdateEvent
{
    PLEvent e;
    membufWindow* window;

    static void* handle(AsyncUpdateEvent*);
    static void destroy(AsyncUpdateEvent*);
};

void* AsyncUpdateEvent::handle(AsyncUpdateEvent* e)
{
    e->window->Update();
    return 0;
}

void AsyncUpdateEvent::destroy(AsyncUpdateEvent* e)
{
    NS_IF_RELEASE(e->window);
    delete e;
}


nsresult membufWindow::InvalidateHelper(PRBool aIsSynchronous)
{
    nsCOMPtr<nsIRegion> oldDirtyRegion(do_CreateInstance(kRegionCID));
    //
    // We keep around the current dirty region so that once it's been
    // cleaned up we can still invalidate all of the children that
    // need to be invalidated.
    //
    oldDirtyRegion->SetTo(*mDirtyRegion.get());

    if(aIsSynchronous) {
        Update();
    } else {
        //
        // Post an Update request into the event queue. Essentially
        // this just means that the redraw is deferred until the
        // current event is done. This can help prevent loops in the
        // rendering code.
        //
        nsresult rv;
        nsCOMPtr<nsIEventQueueService> eventQService = do_GetService(kEventQueueServiceCID, &rv);

        if (NS_FAILED(rv)) {
            NS_WARNING("Could not obtain event queue service");
            return rv;
        }

        nsCOMPtr<nsIEventQueue> eventQueue;
        rv = eventQService->GetThreadEventQueue(NS_CURRENT_THREAD, getter_AddRefs(eventQueue));
        if (NS_FAILED(rv)) {
            NS_WARNING("Could not get the thread event queue");
            return rv;
        }

        AsyncUpdateEvent* aue = new AsyncUpdateEvent;

        PL_InitEvent((PLEvent*)aue, 0,
                     (PLHandleEventProc)AsyncUpdateEvent::handle,
                     (PLDestroyEventProc)AsyncUpdateEvent::destroy);
        aue->window = this;
        NS_IF_ADDREF(aue->window);

        eventQueue->PostEvent((PLEvent*)aue);
    }

    for(nsIWidget* child = GetFirstChild();
        child != 0;
        child = child->GetNextSibling())
    {
        PRBool vis;
	nsRect bbox;
        nsRect rec;
        //
        // Only invalidate this child if it's visible and intersects with
        // our invalidation region. This check could be a little more strict
        // and check on every rectangle in the region iff the child is in
        // the bounding box...
        //
        child->IsVisible(vis);
        child->GetBounds(rec);
        oldDirtyRegion->GetBoundingBox(&bbox.x, &bbox.y, &bbox.width, &bbox.height);
        if(vis && oldDirtyRegion->ContainsRect(rec.x, rec.y, rec.width, rec.height)) {
            child->Invalidate(bbox, aIsSynchronous);
        }
    }
    return NS_OK;
}


nsresult
membufWindow::Paint(nsIRenderingContext& aRenderingContext, nsIRegion* aDirtyRegion)
{
#ifdef PR_LOGGING
    static int depth = 0;
    depth++;
    PR_LOG(nsMembufWindowLog, PR_LOG_DEBUG, ("(0x%x) OnPaint begins; depth = %d", this, depth));
    if( depth > 4 ) NS_ABORT();
#endif
    nsEventStatus eventStatus;
    nsPaintEvent event( true, NS_PAINT, this );
    event.rect = nsnull;
    event.region = aDirtyRegion;
    event.renderingContext = &aRenderingContext;
    event.widget = this;
#ifdef WIN32
    event.time = ::GetMessageTime();
#else
    event.time = PR_Now();
#endif
    event.refPoint.x = 0;
    event.refPoint.y = 0;
    //
    // Dispatch our event, this will cause some painting. This isn't
    // what causes *all* painting, though.
    //
    NS_ADDREF(aDirtyRegion);
    nsresult result = DispatchEvent(&event, eventStatus);
    NS_RELEASE(aDirtyRegion);
#ifdef PR_LOGGING
    depth--;
    PR_LOG(nsMembufWindowLog, PR_LOG_DEBUG, ("(0x%x) membufWindow::OnPaint[ region ]", this));
#endif


    return result;
}

nsresult
membufWindow::Paint(nsIRenderingContext& aRenderingContext, const nsRect& aBounds)
{
    return nsBaseWidget::Paint(aRenderingContext, aBounds);
}

void membufWindow::InitGUIEvent(nsGUIEvent &aEvent, PRUint32 aMsg)
{
    memset(&aEvent, 0, sizeof(nsGUIEvent));
    aEvent.eventStructType = NS_GUI_EVENT;
    aEvent.message = aMsg;
    aEvent.widget = NS_STATIC_CAST(nsIWidget *, this);
}

void membufWindow::DispatchGotFocusEvent(void)
{
    nsGUIEvent event(true,NS_GOTFOCUS,this);
    InitGUIEvent(event, NS_GOTFOCUS);
    nsEventStatus status;
    DispatchEvent(&event, status);
}

void membufWindow::DispatchLostFocusEvent(void)
{
    nsGUIEvent event(true,NS_LOSTFOCUS,this);
    InitGUIEvent(event, NS_LOSTFOCUS);
    nsEventStatus status;
    DispatchEvent(&event, status);
}

void membufWindow::DispatchActivateEvent(void)
{
    nsGUIEvent event(true,NS_ACTIVATE,this);
    InitGUIEvent(event, NS_ACTIVATE);
    nsEventStatus status;
    DispatchEvent(&event, status);
}

gfxASurface*
membufWindow::GetThebesSurface()
{
#ifdef PR_LOGGING
    PR_LOG(nsMembufWindowLog, PR_LOG_DEBUG, ("(0x%x) membufWindow::GetThebesSurface[]", this));
#endif

    if (mParent)
    {
        //
        // To get the parent's surface we iterate through
        // all of our parents up to the top (if we have to)
        // and try and extract a surface. 
        //
        // We have to do this iteration thing because we
        // don't know that any given parent is actually a
        // membufWindow (which is why we use GetNativeData
        // vs. GetThebesSurface, as the latter could return
        // something else).
        //
	nsIWidget* parent = GetParent();
        gfxASurface* surf = 0;
	while( parent && !surf )
        {
            nsIWidget* current(parent);
            parent = current->GetParent();
            if (!parent) surf = current->GetThebesSurface();
        }
        if (!surf) {
#ifdef PR_LOGGING
            PR_LOG(nsMembufWindowLog, PR_LOG_ALWAYS, ("(0x%x) membufWindow::GetThebesSurface FAILED!", this));
#endif
            return nsnull;
        }
	//
	// Get our absolute position into this surface,
	// and set our device offset.
	//
        int xoff, yoff;
	calculateAbsolutePosition (xoff, yoff);
	surf->SetDeviceOffset (xoff, yoff);
	return surf;
    }
    //
    // We are the root window. We need to either allocate
    // a new surface or return the current one if we have
    // one.
    //
    if (mRootSurface)
    {
        mRootSurface->SetDeviceOffset (0,0);
        //
        // We have a surface. If it is of the correct
        // dimensions then we should just set the device
        // offset to zero and return it. If we have
        // resized since the last time we were asked for
        // a surface then we need to make a new one.
        //
        // XXX: Are there any issues with deleting a
        // XXX: surface? Could there be old rendering
        // XXX: contexts around?
        //
        gfxImageSurface* im = (gfxImageSurface*)mRootSurface.get();
        if ((im->Width() == mBounds.width) && (im->Height() == mBounds.height))
        {
#ifdef PR_LOGGING
            PR_LOG(nsMembufWindowLog, PR_LOG_ALWAYS, ("(0x%x) membufWindow::GetThebesSurface[ returning = 0x%x ]", this, mRootSurface.get()));
#endif
            return mRootSurface;
        }
        //
        // Our size has changed. We need to allocate a
        // new surface. We shouldn't have to put an
        // invalidate request in here, because the resize
        // handler should have already done it.
        //
        mRootSurface = 0;
    }
    //
    // Allocate a new surface. The Membuf graphics platform always
    // allocates image buffer surfaces for us.
    //
    mRootSurface = gfxPlatform::GetPlatform()->CreateOffscreenSurface (
        mBounds.width, mBounds.height, gfxASurface::ImageFormatARGB32,
        PR_TRUE);
#ifdef PR_LOGGING
    PR_LOG(nsMembufWindowLog, PR_LOG_ALWAYS, ("(0x%x) membufWindow::GetThebesSurface[ returning new surface = 0x%x; width = %d height = %d ]", this, mRootSurface.get(), mBounds.width, mBounds.height));
#endif
    return mRootSurface;
}
